import inspect

import matplotlib as mpl

from palettable.cartocolors.sequential import *
from palettable.cmocean.sequential import *
from palettable.colorbrewer.sequential import *
from palettable.scientific.sequential import *

_CMAP_QUAL = {
    "acton": Acton_20,
    "davos": Davos_20,
    "devon": Devon_20,
    "oslo": Oslo_20,
    "tokyo": Tokyo_20,
    "nuuk": Nuuk_20,
    "lapaz": LaPaz_20,
    "lajolla": LaJolla_20,
    "imola": Imola_20,
    "blue": Blues_9_r,
    "bugn": BuGn_9_r,
    "gnbu": GnBu_9_r,
    "bupu": BuPu_9_r,
    "green": Greens_9_r,
    "purple": Purples_9_r,
    "orrd": OrRd_9_r,
    "grey_r": Greys_9,
    "gray_r": Greys_9,
    "orange": Oranges_9_r,
    "pubu": PuBu_9_r,
    "pubugn": PuBuGn_9_r,
    "purd": PuRd_9_r,
    "rdpu": RdPu_9_r,
    "red": Reds_9_r,
    "ylgn": YlGn_9_r,
    "ylgnbu": YlGnBu_9_r,
    "ylorbr": YlOrBr_9_r,
    "ylorrd": YlOrRd_9_r,
    "deep": Deep_20_r,
    "dense": Dense_20_r,
    "gray": Gray_20,
    "ice": Ice_20,
    "haline": Haline_20,
    "solar": Solar_20,
    "thermal": Thermal_20,
    "tempo": Tempo_20_r,
    "ocean": Tempo_20_r,
    "speed": Speed_20_r,
    "ocean-green": Speed_20_r,
    "brown": Turbid_20_r,
    "blugrn": BluGrn_7,
    "grnblu": BluGrn_7_r,
    "mint": Mint_7_r,
    "darkmint": DarkMint_7_r,
    "emerald": Emrld_7_r,
    "magenta": Magenta_7_r,
    "teal": Teal_7_r,
    "teal-green": TealGrn_7_r,
    "sunset-dark": SunsetDark_7_r,
    "sunset": Sunset_7_r,
}


# Extra color maps for various purposes like showing RGB channels of an image
_CMAP_EXTRA = {
    "R": mpl.colors.LinearSegmentedColormap.from_list("R", ["#000000", "#FF0000"]),
    "G": mpl.colors.LinearSegmentedColormap.from_list("G", ["#000000", "#00FF00"]),
    "B": mpl.colors.LinearSegmentedColormap.from_list("B", ["#000000", "#0000FF"]),
    "C": mpl.colors.LinearSegmentedColormap.from_list("C", ["#00FFFF", "#FFFFFF"]),
    "M": mpl.colors.LinearSegmentedColormap.from_list("M", ["#FF00FF", "#FFFFFF"]),
    "Y": mpl.colors.LinearSegmentedColormap.from_list("Y", ["#FFFF00", "#FFFFFF"]),
}



import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt

from ._colormap import _CMAP_QUAL, _CMAP_EXTRA

__all__ = [
    "set_context",
    "set_save_context",
    "reset_defaults",
    "set_image",
    "set_scalebar",
]

# TODO implement a set() fuction for all underlying set_*


def set_context(mode="paper", fontfamily="sans-serif", fontweight="normal", rc=None):
    """
    Set context for images with mode, fontfamily and fontweight. Additional,
    rc params can also be passed as dict

    Parameters
    ----------
    mode : str, optional
        Plotting context mode. Depending on the context, axes width, fontsize, layout etc.
        are scaled. Options are 'paper', 'notebook', 'presentation', 'talk' and 'poster',
        by default "paper".
    fontfamily : str, optional
        Font-family to use, by default "sans-serif".
    fontweight : str, optional
        Font-weight to use. Options include 'normal' and 'bold', by default "bold".
    rc : dict, optional
        Additional `matplotlib.rcParams` to be passed to matplotlib, by default None.

    Examples
    --------
        >>> import seaborn_image as isns
        >>> isns.set_context(mode="poster", fontfamily="sans-serif")
        >>> isns.set_context(rc={"axes.edgecolor": "red"})

    """

    if mode == "paper":
        plt.rc("axes", linewidth=1.5)
        plt.rc("axes", titlesize=15, titleweight=fontweight)
        plt.rc("axes", labelsize=15, labelweight=fontweight)
        mpl.rcParams.update({"figure.constrained_layout.wspace": 0.2})
        font = {"family": fontfamily, "weight": fontweight, "size": 10}

    if mode == "notebook" or mode == "presentation" or mode == "talk":
        plt.rc("axes", linewidth=2.5)
        plt.rc("axes", titlesize=20, titleweight=fontweight)
        plt.rc("axes", labelsize=20, labelweight=fontweight)
        mpl.rcParams.update({"figure.constrained_layout.wspace": 0.3})
        font = {"family": fontfamily, "weight": fontweight, "size": 15}

    if mode == "poster":
        plt.rc("axes", linewidth=3.5)
        plt.rc("axes", titlesize=25, titleweight=fontweight)
        plt.rc("axes", labelsize=25, labelweight=fontweight)
        mpl.rcParams.update({"figure.constrained_layout.wspace": 0.4})
        font = {"family": fontfamily, "weight": fontweight, "size": 20}

    plt.rc("font", **font)

    if rc is not None:
        assert isinstance(rc, dict), "'rc' must be a dict"
        mpl.rcParams.update(rc)


def set_save_context(dpi=300):
    """
    Set dpi for saving figures to disk

    Parameters
    ----------
    dpi : int, optional
        Image dpi for saving, by default 300.

    Examples
    --------
        >>> import seaborn_image as isns
        >>> isns.set_save_context(dpi=200)
    """
    plt.rc("savefig", dpi=dpi, bbox="tight")


def reset_defaults():
    """
    Reset rcParams to matplotlib defaults

    Examples
    --------
        >>> import seaborn_image as isns
        >>> isns.reset_defaults()
    """
    mpl.rcParams.update(mpl.rcParamsDefault)


def set_image(cmap="deep", origin="upper", interpolation="nearest", despine=False):
    """
    Set deaults for plotting images

    Parameters
    ----------
    cmap : str, optional
        Colormap to use accross images, by default to "deep".
    origin : str, optional
        Image origin - same as in `matplotlib.pyplot.imshow`, by default "upper".
    interpolation : str, optional
        Image interpolation - same as in `matplotlib.pyplot.imshow`, by default "nearest".
    despine : bool, optional
        Despine image and colorbar axes, by default False.

    Examples
    --------
        >>> import seaborn_image as isns
        >>> isns.set_image(cmap="inferno", interpolation="bicubic")
        >>> isns.set_image(despine=False)

    """

    # Load colormap from palletable
    if isinstance(cmap, str) and cmap in _CMAP_QUAL.keys():
        cmap_qual_mpl = _CMAP_QUAL.get(cmap).mpl_colormap
        if cmap not in mpl.colormaps.keys():
            mpl.colormaps.register(name=cmap, cmap=cmap_qual_mpl)

    # Load extra colormap
    if isinstance(cmap, str) and cmap in _CMAP_EXTRA.keys():
        cmap_channel_mpl = _CMAP_EXTRA.get(cmap)
        if cmap not in mpl.colormaps.keys():
            mpl.colormaps.register(name=cmap, cmap=cmap_channel_mpl)

    # change the axes spines
    # "not" is required because of the despine parameter name
    # if depine is True ---> you don't want the axes spines
    # and therfore, axes.spines.bottom == False (or not True)
    mpl.rcParams.update(
        {
            "axes.spines.bottom": not despine,
            "axes.spines.left": not despine,
            "axes.spines.right": not despine,
            "axes.spines.top": not despine,
        }
    )

    plt.rc("image", cmap=cmap, origin=origin, interpolation=interpolation)


def set_scalebar(
    color="white",
    location="lower right",
    width_fraction=0.025,
    height_fraction=None,
    length_fraction=0.3,
    scale_loc="top",
    box_alpha=0,
    rc=None,
):
    """Set scalebar properties such as color, scale_loc,
    height_fraction, length_fraction, box_alpha, etc.
    To pass more properties that are not specified as key word
    argument, use the `rc` parameter. Refer to https://github.com/ppinard/matplotlib-scalebar
    for more information on additional parameters.

    Parameters
    ----------
        color : str, optional
            Color of the scalebar, by default "white".
        location : str, optional
            Scalebar location on the image (same as `matplotlib` legend).
            by default "lower right".
        width_fraction : float, optional
            By default 0.025.
        height_fraction : float, optional
            Deprecated - use `width_fraction` instead.
        length_fraction : float, optional
            By default 0.3
        scale_loc :str, optional
            Location of the scale number and units with respect to
            the bar, by default "top".
        box_alpha : float, optional
            Transparency of the box that contains
            the scalebar artist, by default 0.
        rc : dict, optional
            Dictionary of scalebar properties to be set, by default None.

    Examples
    --------
        >>> import seaborn_image as isns
        >>> isns.set_scalebar(color = "red")
        >>> isns.set_scalebar(scale_loc = "bottom")

    """
    if height_fraction:
        warnings.warn(
            "`height_fraction` is deprecated. Use `width_fraction` instead.",
            DeprecationWarning,
        )
        width_fraction = height_fraction

    mpl.rcParams.update({"scalebar.color": color})
    mpl.rcParams.update({"scalebar.width_fraction": width_fraction})
    mpl.rcParams.update({"scalebar.length_fraction": length_fraction})
    mpl.rcParams.update({"scalebar.scale_loc": scale_loc})
    mpl.rcParams.update({"scalebar.location": location})
    mpl.rcParams.update({"scalebar.box_alpha": box_alpha})

    if rc is not None:
        assert isinstance(rc, dict), "'rc' must be a dict"
        for key, value in rc.items():
            mpl.rcParams.update({f"scalebar.{key}": value})



import matplotlib as mpl
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib_scalebar.scalebar import ScaleBar
from mpl_toolkits.axes_grid1 import axes_size, make_axes_locatable

from ._colormap import _CMAP_QUAL, _CMAP_EXTRA
from .utils import despine, scientific_ticks

# dimensions for scalebar
_DIMENSIONS = {
    "si": "si-length",
    "si-reciprocal": "si-length-reciprocal",
    "imperial": "imperial-length",
    "angle": "angle",
    "pixel": "pixel-length",
}


class _SetupImage(object):
    def __init__(
        self,
        data,
        ax=None,
        cmap=None,
        vmin=None,
        vmax=None,
        alpha=None,
        origin=None,
        interpolation=None,
        norm=None,
        robust=False,
        perc=None,
        diverging=False,
        dx=None,
        units=None,
        dimension=None,
        cbar=None,
        orientation="v",
        cbar_label=None,
        cbar_ticks=None,
        showticks=False,
        despine=None,
        extent=None,
    ):

        self.data = data
        self.ax = ax
        self.cmap = cmap
        self.vmin = vmin
        self.vmax = vmax
        self.alpha = alpha
        self.origin = origin
        self.interpolation = interpolation
        self.norm = norm
        self.robust = robust
        self.diverging = diverging
        self.perc = perc
        self.dx = dx
        self.units = units
        self.dimension = dimension
        self.cbar = cbar
        self.orientation = orientation
        self.cbar_label = cbar_label
        self.cbar_ticks = cbar_ticks
        self.showticks = showticks
        self.despine = despine
        self.extent = extent

    def _setup_figure(self):
        """Wrapper to setup image with the desired parameters"""
        if self.ax is None:
            f, ax = plt.subplots()
        else:
            f = self.ax.get_figure() or plt.gcf()
            ax = self.ax

        return f, ax

    def _setup_scalebar(self, ax):
        """Setup scalebar for the image"""

        if self.dx:
            dx = self.dx
            if self.units:
                units = self.units
            else:
                raise AttributeError(
                    "'units' must be specified when 'dx' (scalebar) is used"
                )

        if self.dimension is None:
            _dimension = _DIMENSIONS.get("si")
        elif self.dimension in _DIMENSIONS.keys():
            _dimension = _DIMENSIONS.get(self.dimension)
        else:
            raise ValueError(
                f"Unsupported dimension. Supported dimensions are : {_DIMENSIONS.keys()}"
            )

        scalebar = ScaleBar(dx=dx, units=units, dimension=_dimension)

        ax.add_artist(scalebar)

    def plot(self):
        f, ax = self._setup_figure()

        if isinstance(self.cmap, str) and self.cmap in _CMAP_QUAL.keys():
            self.cmap = _CMAP_QUAL.get(self.cmap).mpl_colormap

        if isinstance(self.cmap, str) and self.cmap in _CMAP_EXTRA.keys():
            self.cmap = _CMAP_EXTRA.get(self.cmap)

        if self.robust:
            min_robust = False
            max_robust = False
            if self.vmin is None:
                # remember that vmin was None and now set to new value
                min_robust = True
                self.vmin = np.nanpercentile(self.data, self.perc[0])
            if self.vmax is None:
                # remember that vmax was None and now set to new value
                max_robust = True
                self.vmax = np.nanpercentile(self.data, self.perc[1])

        if self.diverging:
            # Force vmin to have the same absolute value as vmax so that 0 is in the middle.

            if self.vmax is None and self.vmin is None:
                self.vmax = np.abs(self.data).max()
                self.vmin = -self.vmax
            elif self.vmax is None:
                self.vmax = abs(self.vmin)
            elif self.vmin is None:
                self.vmin = -abs(self.vmax)
            else:
                self.vmax = max(abs(self.vmin), abs(self.vmax))
                self.vmin = -self.vmax

        if self.norm == "cbar_log":
            self.norm = colors.LogNorm(vmin=self.vmin, vmax=self.vmax)

        # TODO move everything other than data to kwargs
        _map = ax.imshow(
            self.data,
            cmap=self.cmap,
            vmin=self.vmin,
            vmax=self.vmax,
            origin=self.origin,
            alpha=self.alpha,
            interpolation=self.interpolation,
            norm=self.norm,
            extent=self.extent,
        )

        if self.dx:
            self._setup_scalebar(ax)

        if self.cbar:
            divider = make_axes_locatable(ax)

            if self.orientation in ["vertical", "v"]:
                self.orientation = "vertical"  # plt.colorbar doesn't take 'v'
                width = axes_size.AxesY(ax, aspect=1.0 / 20)
                pad = axes_size.Fraction(0.5, width)
                cax = divider.append_axes("right", size=width, pad=pad)

            elif self.orientation in ["horizontal", "h"]:
                self.orientation = "horizontal"  # plt.colorbar doesn't take 'h'
                width = axes_size.AxesX(ax, aspect=1.0 / 20)
                pad = axes_size.Fraction(0.5, width)
                cax = divider.append_axes("bottom", size=width, pad=pad)

            else:
                raise ValueError(
                    "'orientation' must be either : 'horizontal' or 'h' / 'vertical' or 'v'"
                )

            # extend specific colorbar regions if robust is True
            if self.robust:
                if min_robust and max_robust:
                    cb = f.colorbar(
                        _map, cax=cax, orientation=self.orientation, extend="both"
                    )
                elif min_robust:
                    cb = f.colorbar(
                        _map, cax=cax, orientation=self.orientation, extend="min"
                    )
                elif max_robust:
                    cb = f.colorbar(
                        _map, cax=cax, orientation=self.orientation, extend="max"
                    )
            else:
                cb = f.colorbar(_map, cax=cax, orientation=self.orientation)

            if self.despine is None:
                # if depsine is None, use global settings
                # if all image axes spines are despined, cbar should also be despined
                if (
                    mpl.rcParams["axes.spines.top"] is False
                    and mpl.rcParams["axes.spines.bottom"] is False
                    and mpl.rcParams["axes.spines.left"] is False
                    and mpl.rcParams["axes.spines.right"] is False
                ):
                    self.despine = True

            if self.despine is True:
                cb.outline.set_visible(False)  # remove colorbar outline border
            else:
                self.despine = (
                    False  # NOTE this is only being set inside self.cbar block
                )

            # display only 3 tick marks on colorbar
            if self.orientation in ["vertical"]:
                cax.yaxis.set_major_locator(plt.MaxNLocator(3))
                cax.tick_params(
                    axis="y", width=0, length=0, which="both"
                )  # remove major and minor ytick

            if self.orientation in ["horizontal"]:
                cax.xaxis.set_major_locator(plt.MaxNLocator(3))
                cax.tick_params(
                    axis="x", width=0, length=0, which="both"
                )  # remove major and minor xticks

            # TODO add option for scientific ticks as part of inbuilt functions
            # scientific_ticks(ax=cax)

            if self.cbar_ticks is not None:
                cb.set_ticks(self.cbar_ticks)

            if self.cbar_label is not None:
                cb.set_label(self.cbar_label)

        else:
            cax = None

        if not self.showticks:
            ax.set_yticks([])
            ax.set_xticks([])
            ax.set_ylabel("")
            ax.set_xlabel("")
            # ax.get_yaxis().set_visible(False)
            # ax.get_xaxis().set_visible(False)

        # This block is for handling local despine state
        # If the state doesn't match the global state,
        # local despine state will be preferred
        if self.despine:
            despine(ax=ax, which="all")

        elif self.despine is False:  # if despine is False
            # TODO implement a respine function (?)
            for spine in ["top", "bottom", "right", "left"]:
                ax.spines[spine].set_visible(True)

        return f, ax, cax



import numpy as np
import pooch
from skimage import color, data, io

__all__ = ["load_image"]


POOCH = pooch.create(
    # Use the default cache folder for the OS
    path=pooch.os_cache("seaborn-image"),
    # The remote data is on Github
    base_url="https://github.com/SarthakJariwala/seaborn-image/raw/master/data/",
    # The registry specifies the files that can be fetched
    registry={
        # The registry is a dict with file names and their SHA256 hashes
        "PolymerImage.txt": "7b6798865080adf3ecf11e342f3d86d7b52ea0700020a1f062544ee825fb8a0e",
        "Perovskite.txt": "3228eeade5afec3c2b1ed116b2d4fe35877224d2d9bf7b4a17e04a432e6135c5",
        "cells.tif": "2120cfe08e0396324793a10a905c9bbcb64b117215eb63b2c24b643e1600c8c9",
        "cifar10.npy": "c0a12085b3b82f4a6d1f95e609a40701648a137eb9ff1fb5751071f54cc8e05c",
    },
)


def load_image(name):
    """Load image data shippped with seaborn-image.

    Parameters
    ----------
    name : str
        Name of the image dataset

    Raises
    ------
    ValueError
        If the name of the dataset specified doesn't exist

    Returns
    -------
    `numpy.ndarray`
        Image data as a `numpy` array

    Examples
    --------
    >>> import seaborn_image as isns
    >>> img = isns.load_image("polymer")
    """

    if name == "polymer":
        path = POOCH.fetch("PolymerImage.txt")
        img = np.loadtxt(path, skiprows=1)
        img = img * 1e9  # convert height data from m to nm

    elif name == "polymer outliers":
        path = POOCH.fetch("PolymerImage.txt")
        img = np.loadtxt(path, skiprows=1)
        img = img * 1e9  # convert height data from m to nm
        img[25, 25] = 80  # assign an outlier value to a random pixel

    elif name == "fluorescence":
        path = POOCH.fetch("Perovskite.txt")
        img = np.loadtxt(path)

    elif name == "cells":
        path = POOCH.fetch("cells.tif")
        img = io.imread(path).T

    elif name == "retina-gray":
        img = color.rgb2gray(data.retina())[300:700, 700:900]

    elif name == "cifar10":
        path = POOCH.fetch("cifar10.npy")
        img = np.load(path)

    elif name == "cifar10 list":
        path = POOCH.fetch("cifar10.npy")
        img_array = np.load(path)
        img = [arr for arr in img_array]

    else:
        raise ValueError(
            f"No '{name}' image dataset. Available image datasets include: polymer, polymer outliers, fluorescence, cells, retina-gray, cifar10, cifar list"
        )

    return img
